/*
  Copyright 2020 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth.network

import com.kakao.sdk.auth.AuthApiClient
import com.kakao.sdk.auth.TokenManagerProvider
import com.kakao.sdk.common.Constants
import com.kakao.sdk.common.model.ApiErrorCause
import com.kakao.sdk.common.model.ApiErrorResponse
import com.kakao.sdk.common.util.KakaoJson
import okhttp3.Interceptor
import okhttp3.Request
import okhttp3.Response
import okhttp3.ResponseBody

/**
 * @suppress
 *
 * API 요청에 AccessToken을 추가하는 인터셉터
 * -401 발생시 자동 갱신
 */
class AccessTokenInterceptor(
    private val tokenManagerProvider: TokenManagerProvider = TokenManagerProvider.instance,
    private val authApiClient: AuthApiClient = AuthApiClient.instance
) : Interceptor {
    override fun intercept(chain: Interceptor.Chain): Response {
        val usedAccessToken = tokenManagerProvider.manager.getToken()?.accessToken

        val request =
            usedAccessToken?.let {
                requestWithAuthorization(chain.request(), it)
            } ?: chain.request()

        val originalResponse = chain.proceed(request)

        if (originalResponse.isSuccessful) {
            return originalResponse
        }

        val body = originalResponse.body()
        val bodyString = body?.string()

        // 에러 응답 파싱을 위해 body를 한번 소비했기 때문에 새로 만들어주지 않으면 다른 인터셉터에서 body 접근 시 크래시 난다.
        val response = originalResponse.newBuilder()
            .body(ResponseBody.create(body?.contentType(), bodyString))
            .build()

        val apiErrorCause = bodyString?.let {
            KakaoJson.fromJson<ApiErrorResponse>(it, ApiErrorResponse::class.java)
        }?.let {
            KakaoJson.fromJson<ApiErrorCause>(it.code.toString(), ApiErrorCause::class.java)
        }

        if (apiErrorCause == ApiErrorCause.InvalidToken) {

            // 나중에 들어온 요청들 pending
            synchronized(this) {

                // resume 돼서 들어왔을 때 현재 토큰 보고
                val currentToken = tokenManagerProvider.manager.getToken()
                if (currentToken != null ) {

                    // 이전 요청에서 넣었던 토큰과 현재 토큰이 다르면
                    if (currentToken.accessToken != usedAccessToken) {
                        // 이미 앞의 요청에서 갱신됐다고 판단하고 현재 토큰으로 재요청
                        return chain.proceed(requestWithAuthorization(request, currentToken.accessToken))
                    }
                    else {
                        try {
                            // 갱신 후 재요청
                            val newToken = authApiClient.refreshAccessToken(currentToken)
                            return chain.proceed(requestWithAuthorization(request, newToken.accessToken))
                        } catch (e: Throwable) {
                            throw java.io.IOException(e)
                        }
                    }
                }
            }
        }

        return response
    }

    companion object {
        fun requestWithAuthorization(request: Request, accessToken: String): Request =
            request.newBuilder()
                .removeHeader(Constants.AUTHORIZATION)
                .addHeader(Constants.AUTHORIZATION, "${Constants.BEARER} $accessToken")
                .build()
    }
}
